/* 
 * nngfs/super.c
 * 
 * Copyright (C) 2008
 * Manish Katiyar (mkatiyar@gmail.com) 
 *
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>
#include <linux/slab.h>
#include <linux/mpage.h>
#include <linux/statfs.h>
#include <linux/init.h>

#include "nngfs.h"

#define NNGFS_MNT_FORCE 0x1
#define MOD_AUTHOR "Manish Katiyar"
#define MODULE_DESC "A Simple disk based filesystem"

enum {
Opt_force,
Opt_err
};

static match_table_t tokens = {
{Opt_force,"force"},
{Opt_err,NULL}
};

static struct kmem_cache *nngfs_inode_info_p;
extern struct file_operations nngfs_dir_operations;
extern struct file_operations nngfs_file_operations;
extern struct inode_operations nngfs_dir_inode_operations;
extern struct inode_operations nngfs_file_inode_operations;
extern struct address_space_operations nngfs_aops;

static inline int get_opt(int val, int opt){
       return (val)&(opt);
}

static inline void set_opt(int *val, int opt){
       *val = (*val)|(opt);
       return;
}

static int parse_options(char *options, struct nngfs_info *nsi){
       char *p;
       substring_t args[MAX_OPT_ARGS];

       if(!options)
          return 1;

       while ((p = strsep (&options, ",")) != NULL) {
        int token;
         if (!*p)
          continue;
         token = match_token(p, tokens, args);
         switch (token) {
            case Opt_force : set_opt(&(nsi->mount_opt),NNGFS_MNT_FORCE);
                        break;
            default : printk(KERN_CRIT "NNGFS : Unknown mount option - %s\n",p);
                     return 0;
         };
        };
        return 1;
}

void dump_inode (struct nngfs_inode *ptr){
	printk(KERN_INFO "dumping inode data for inode [%d] :\n",ptr->inode);
	printk(KERN_INFO " blockno = %d, filetype = 0x%x,uid = %d, gid = %d, size = %d, perm=%x\n",ptr->blockno,
	                                                                                           ptr->filetype,
												   ptr->uid,
												   ptr->gid,
												   ptr->size,
												   ptr->perm);
	return;
}

struct nngfs_inode *nngfs_get_inode(struct super_block *sb,unsigned32 ino,struct buffer_head **bhp){
	struct buffer_head *bh;
        /* Our inode's data is at block 1. block 0 contains the superblock
         * Rest all the blocks contain data specific to inode. They have one to one mapping
  	 */
	if(ino > (sizeof(ino)*8-1)){
	 goto invalid;
	}
        if(!(bh=sb_bread(sb,NNGFS_INODE_BLOCK))){
#ifdef DEBUG
	printk(KERN_INFO "NGFS (%s) : Unable to read block number [%d] \n",__FUNCTION__,NNGFS_INODE_BLOCK);
#endif
	goto invalid;
	}
        *bhp = bh;
#ifdef DEBUG
	dump_inode((struct nngfs_inode *)((char *)bh->b_data));
#endif
	return (struct nngfs_inode *)((char *)bh->b_data+sizeof(struct nngfs_inode)*(ino-1));
invalid : 
	return (ERR_PTR(-EIO));
}


void nngfs_read_inode(struct inode *inode){	
	unsigned32 ino = inode->i_ino;
	struct buffer_head *bh;
	struct nngfs_inode *ninode=NULL;
	ninode = nngfs_get_inode(inode->i_sb,ino,&bh); /* This is the raw inode read from disk */
#ifdef DEBUG
	dump_inode(ninode);
#endif
	inode->i_private = ninode;
        inode->i_mode = ninode->filetype;
	inode->i_size = ninode->size;
	inode->i_uid = ninode->uid;
	inode->i_gid = ninode->gid;
        inode->i_atime = ninode->atime;
        inode->i_mtime = ninode->mtime;
        inode->i_ctime = ninode->ctime;
	if(S_TYPE(ninode->filetype,S_IFREG)){
	inode->i_fop = &nngfs_file_operations;
	inode->i_op = &nngfs_file_inode_operations;
        inode->i_mapping->a_ops = &nngfs_aops;
 	}
	else{
	if(S_TYPE(ninode->filetype,S_IFDIR)){
	inode->i_fop = &nngfs_dir_operations;
	inode->i_op = &nngfs_dir_inode_operations;
        inode->i_mapping->a_ops = &nngfs_aops;
	}
	else
	 printk(KERN_INFO "NGFS (%s) : Inode number [%d] is a unknown file\n",__FUNCTION__,ino);
	}
	return;
}

void nngfs_destroy_inode(struct inode *inode){
	kmem_cache_free(nngfs_inode_info_p,NNGFS_I(inode));
	return;
}

int nngfs_write_inode(struct inode *ino, int wait){
        unsigned32 inum = ino->i_ino;
        struct nngfs_inode nino ;
        struct buffer_head *bh;
        memset(&nino,0,sizeof nino);
        bh = sb_bread(ino->i_sb,NNGFS_INODE_BLOCK);
        if(!bh){
#ifdef DEBUG
        printk(KERN_INFO "NGFS (%s) : Could not read buffer for block num [%d] inode [%d] \n",__FUNCTION__,inum,inum);
#endif
        return -1;
        }

        nino.inode = ino->i_ino;
        nino.filetype = ino->i_mode;
        nino.atime = ino->i_atime;
        nino.ctime = ino->i_ctime;
        nino.mtime = ino->i_mtime;
        nino.blockno = ino->i_ino+1;
        nino.uid = ino->i_uid;
        nino.gid = ino->i_gid;
        nino.size = ino->i_size;
        nino.magic = 0xdeadbeef;
        memcpy((char *)bh->b_data+(ino->i_ino-1)*sizeof(struct nngfs_inode),&nino,sizeof(nino));
        mark_buffer_dirty(bh);
	return 0;
}

struct inode *nngfs_alloc_inode(struct super_block *sb){
        struct nngfs_inode_info *nsi = kmem_cache_alloc(nngfs_inode_info_p,GFP_KERNEL);
	if(!nsi){
#ifdef DEBUG
        printk(KERN_INFO "NGFS (%s) : Could not allocate inode, returning NULL \n",__FUNCTION__);
#endif
	return NULL;
	}
       // memset(nsi,0,sizeof(struct nngfs_inode_info));
	nsi->vino.i_version=1;
        return &(nsi->vino);
}


void nngfs_put_super(struct super_block *sb){
        struct nngfs_info * nsb = (struct nngfs_info *)(sb->s_fs_info);
        struct buffer_head *bh = nsb->sbh;
        struct nngfs_superblock *sngfs = nsb->ngfs_sb;
        sngfs->dirty = NNGFS_CLEAN;
#ifdef DEBUG
        printk(KERN_INFO "NGFS (%s) : Cleaning the dirty bit \n",__FUNCTION__);
#endif
        mark_buffer_dirty(bh);
        brelse(bh);
        kfree(nsb);
	return;
}

static unsigned int nngfs_blocks(struct nngfs_superblock *sngfs){
    unsigned int bitmap = sngfs->inode_bitmap;
    unsigned int count=0,i=0;
    for(i=0;i<sizeof(bitmap)*8;i++)
            if(bitmap & (1<<i))
                  count++;
    return count+2 /* 2 more than the number of inodes. For sb and inodeinfo */;
}

static int nngfs_statfs(struct dentry *dentry,struct kstatfs *stats){
  struct super_block *sb = dentry->d_sb;
  struct nngfs_info * nsb = (struct nngfs_info *)(sb->s_fs_info);
  struct nngfs_superblock *sngfs = nsb->ngfs_sb;

  stats->f_type = NNGFS_MAGIC;
  stats->f_bsize = sb->s_blocksize;
  stats->f_namelen = NNGFS_NAME_LEN;
  stats->f_blocks = stats->f_files = sizeof(sngfs->inode_bitmap)*8;
  stats->f_bavail = (stats->f_blocks - nngfs_blocks(sngfs));
  stats->f_bfree = stats->f_ffree = stats->f_bavail;
  return 0;
}

void nngfs_write_super(struct super_block *sb){
        struct nngfs_info * nsb = (struct nngfs_info *)(sb->s_fs_info);
        struct buffer_head *bh = nsb->sbh;
        struct nngfs_superblock *sngfs = nsb->ngfs_sb;
        sngfs->dirty = NNGFS_DIRTY;
        mark_buffer_dirty(bh);
        sb->s_dirt = 0;
	return;
}

struct super_operations nngfs_super_operations = {
        .alloc_inode = nngfs_alloc_inode,
        .destroy_inode = nngfs_destroy_inode,
        .read_inode = nngfs_read_inode,
        .write_inode = nngfs_write_inode,
        .put_super = nngfs_put_super,
        .write_super = nngfs_write_super,
        .statfs = nngfs_statfs,
};


static int nngfs_fill_super(struct super_block *sb,void *data,int silent){
	struct buffer_head *bh=0;
	struct nngfs_info *nsi=NULL;
	struct nngfs_superblock *sngfs;
	struct inode *root=NULL;
	unsigned32 blocksize = sb->s_blocksize; /* Since we will be using a loopback device, this should be 4096 - default */
	nsi = kzalloc(sizeof(*nsi),GFP_KERNEL);
	if(!nsi){
#ifdef DEBUG
	printk(KERN_INFO "NGFS (%s) : Failed to allocate memory \n",__FUNCTION__);
#endif
	  return -ENOMEM;
	}
	sb->s_fs_info = nsi;
	blocksize = 4096 ;
	if(!sb_set_blocksize(sb,blocksize)){
#ifdef DEBUG
         printk(KERN_INFO "NGFS (%s) : Bad blocksize \n",__FUNCTION__);
#endif
         return -EACCES;
        }
	if(!(bh=sb_bread(sb,NNGFS_SUPERBLOCK))){
#ifdef DEBUG
	printk(KERN_INFO "NGFS (%s) : Failed to read buffer [%p]\n",__FUNCTION__,bh);
#endif
	   goto failed;
        }
	sngfs = (struct nngfs_superblock *)(bh->b_data);
	nsi->ngfs_sb = sngfs;
	sb->s_magic = sngfs->magic;

        nsi->mount_opt = 0;
        if (!parse_options ((char *) data, nsi))
             goto failed;

	if(sb->s_magic != NNGFS_MAGIC){
#ifdef DEBUG
	printk(KERN_INFO "NGFS (%s) : Bad magic number found \n",__FUNCTION__);
#endif
	goto failed;
	}
	if(!get_opt(nsi->mount_opt,NNGFS_MNT_FORCE) && sngfs->dirty != NNGFS_CLEAN){
#ifdef DEBUG
	printk(KERN_CRIT "NGFS (%s) : File system not clean due to previous dirty unmount. Use \"-o force\" option to force mount. \n",__FUNCTION__);
#endif
	goto failed;
	}

        sngfs->dirty = NNGFS_DIRTY;
	nsi->sbh = bh;
	sb->s_op = &nngfs_super_operations;

	root=iget(sb,NNGFS_ROOT_INODE);
        if (!root || is_bad_inode(root))
          goto failed;
	sb->s_root = d_alloc_root(root);
        if(!sb->s_root)
          goto failed;
	root->i_op = &nngfs_dir_inode_operations;
	root->i_fop = &nngfs_dir_operations;
       // root->i_mapping->a_ops = &nngfs_aops;
        nngfs_write_super(sb);
	return 0;
failed :
        kfree(nsi);
	return -EINVAL;
}

static int nngfs_get_sb(struct file_system_type *fs,int flags,
                           const char *dev_name, void *data, struct vfsmount *mnt){
	 return get_sb_bdev(fs,flags, dev_name,data,nngfs_fill_super,mnt);
}

struct file_system_type nngfs_fs={
        .owner      = THIS_MODULE,
 	.name       = "nngfs",
	.get_sb     = nngfs_get_sb,
	.kill_sb    = kill_block_super,
	.fs_flags   = FS_REQUIRES_DEV,
};

static void init_once(void * foo, struct kmem_cache * cachep, unsigned long flags)
{
 struct nngfs_inode_info *ei = (struct nngfs_inode_info *) foo; 
 inode_init_once(&ei->vino);
}

static unsigned32 init_nngfs_inodecache(void){
	nngfs_inode_info_p = kmem_cache_create("nngfs_inode_cache",sizeof(struct nngfs_inode_info),
                                                0,(SLAB_RECLAIM_ACCOUNT|SLAB_MEM_SPREAD),init_once);
	if(!nngfs_inode_info_p)
	   return -ENOMEM;
	return 0;
}

static int nngfs_fs_init(void){
	unsigned32 err;
	err=register_filesystem(&nngfs_fs);
	err=init_nngfs_inodecache();
        if(err)
           goto out;
	return err;
out:
	kmem_cache_destroy(nngfs_inode_info_p);
	return err;
}

static void nngfs_fs_exit(void){
	unregister_filesystem(&nngfs_fs);
	kmem_cache_destroy(nngfs_inode_info_p);
        return ;
}

module_init(nngfs_fs_init);
module_exit(nngfs_fs_exit);

MODULE_AUTHOR(MOD_AUTHOR);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(MODULE_DESC);

